import scanpy as sc
import anndata as ad
import numpy as np
import pandas as pd
import squidpy as sq

import scipy
import random 
import seaborn as sns
import math
from sklearn.preprocessing import  OneHotEncoder
from scipy.sparse import csr_matrix

import os
import os.path as osp
import sys
from collections import Counter

import numba
import multiprocessing as mp
from functools import partial
import torch as t
import torch.distributions as dists
import argparse as arp
from typing import Dict,Callable,List


###############################################
#RCTD
def assign_random_cellnum(cellnum,celtypenum):
    node = list(np.random.choice(np.arange(cellnum),(celtypenum-1),replace = True))  #改为True
    node.append(cellnum)
    node.sort() #n个节点排序
    b=np.zeros(celtypenum,dtype="int32")
    for i in range(celtypenum):
        if i == 0:
            b[i] = node[i]
        else:
            b[i] = node[i]-node[i-1]      
    return b

def generat_grid(row=50, col=50):
    x = np.linspace(0,100,row)
    y = np.linspace(0,100,col)
    x1, x2 = np.meshgrid(x, y)
    spatial = np.array([x1.ravel(), x2.ravel()]).T  # coordinate of x and y
    return x1,x2, spatial

def RCTD_naive(scdata, row=50, col=50, Min = 5, Max = 10, ctkey = 'celltype', maxctnum = 3):
    '''
    scdata: single-cell data
    Num_sample: the number of spot. can be sqrt.
    Min/Max: min/max cell number in each spot
    ctkey: celltype key in adata
    maxctnum: celltype number in each spot
    '''
    #spatial coor
    x1,x2,spatial = generat_grid(row, col) #空间坐标
    
    Num_sample=row * col  #spot数
    Num_pool = scdata.X.shape[0]  #单细胞数 (修改)
    Num_gene = scdata.X.shape[1]  #gene数 （修改）
    celltypes = scdata.obs[ctkey] #各cell的ct
    #Allcelltype = np.array(scdata.obs.CellType.cat.categories) #提取celltype name
    Allcelltype = scdata.obs[ctkey].unique() #提取celltype name
    Num_celltype = Allcelltype.shape[0] #celltype数
    #onehot
    onehot_encoder = OneHotEncoder(sparse=True)
    integer_encoded = np.array(celltypes).reshape(len(celltypes), 1)
    cell2celltype = onehot_encoder.fit_transform(integer_encoded)
    
    #assigncelltospot
    
    cellnum = np.random.randint(Min, Max,Num_sample)
    ####
    if Max == 1:
        cellnum = np.ones(Num_sample)
    eachcellnum = np.zeros((Num_sample,maxctnum),dtype = "int32")
    indices = []

    for i in range(Num_sample):
        
        if Max == 1:
            eachcellnum[i,np.random.randint(0,maxctnum,1)] = 1
        else:
            eachcellnum[i,:] = assign_random_cellnum(cellnum[i], maxctnum)
        if Num_celltype>maxctnum:
            celltype = np.random.choice(Allcelltype,maxctnum,replace = True) #抽3个不重复ct  修改为True
        else:
            celltype = Allcelltype
        for j in range(min(maxctnum,celltype.shape[0])):
            choicepool = np.where((scdata.obs[ctkey] == celltype[j]))[0]  #3个ct对应cell的序号
            tempchoice = np.random.choice(choicepool, eachcellnum[i,j],replace = True)
            indices = np.concatenate((indices,tempchoice))

    indptr = np.concatenate(([0],np.cumsum(cellnum)))
    elements = np.ones(indices.shape[0])
    # 矩阵中第i行非零元素的列号为indices[indptr[i]:indptr[i+1]]，相应的值为elements[indptr[i]:indptr[i+1]]
    spots2cell = csr_matrix((elements, indices, indptr), shape=(Num_sample, Num_pool))
    # 比例矩阵
    W = spots2cell @ cell2celltype
    # 每个spot的表达矩阵
    exp_spots2gene = spots2cell @ scdata.X
    return W, exp_spots2gene,spatial,spots2cell

@numba.jit(nopython=True, parallel=True)
def get_bin_edges(a, bins):
    bin_edges = np.zeros((bins+1,), dtype=np.float64)
    a_min = a.min()
    a_max = a.max()
    delta = (a_max - a_min) / bins
    for i in numba.prange(bin_edges.shape[0]):
        bin_edges[i] = a_min + i * delta

    bin_edges[-1] = a_max  # Avoid roundoff error on last point
    return bin_edges

# Modified from https://github.com/numba/numba-examples
@numba.jit(nopython=True, parallel=False)
def compute_bin(x, bin_edges):
    # assuming uniform bins for now
    n = bin_edges.shape[0] - 1
    a_max = bin_edges[-1]
    # special case to mirror NumPy behavior for last bin
    if x == a_max:
        return n - 1 # a_max always in last bin
    bin = np.searchsorted(bin_edges, x)-1
    if bin < 0 or bin >= n:
        return None
    else:
        return bin
    
# Modified from https://github.com/numba/numba-examples
@numba.jit(nopython=True, parallel=False)
def numba_histogram(a, bin_edges):
    hist = np.zeros((bin_edges.shape[0] - 1,), dtype=np.intp)
    for x in a.flat:
        bin = compute_bin(x, bin_edges)
        if bin is not None:
            hist[int(bin)] += 1
    return hist, bin_edges

# Modified from https://github.com/numba/numba-examples
# Modified from https://rdrr.io/bioc/scRecover/src/R/countsSampling.R
# Downsample cell reads to a fraction
@numba.jit(nopython=True, parallel=True)
def downsample_cell(cell_counts,max_counts):
    readsGet = np.sort(np.random.choice(np.arange(np.sum(cell_counts)), np.intp(max_counts), replace=False))
    cumCounts = np.concatenate((np.array([0]),np.cumsum(cell_counts)))
    counts_new = numba_histogram(readsGet,cumCounts)[0]
    counts_new = counts_new.astype(np.float32)
    return counts_new

def downsample_matrix_by_cell(matrix,max_counts,n_cpus=None):
    if n_cpus is not None:
        with mp.Pool(16) as p:
            matrix_ds = p.map(partial(downsample_cell,max_counts), matrix)
    else:
        matrix_ds = [downsample_cell(i,max_counts) if i.sum() > max_counts else i for i in matrix ]
    return np.array(matrix_ds)

def generat_grid2(locations = [50,50]):
    x = np.linspace(0,100,locations[0])
    y = np.linspace(0,100,locations[1])
    x1, x2 = np.meshgrid(x, y)
    spatial = np.array([x1.ravel(), x2.ravel()]).T  # coordinate of x and y
    return x1,x2, spatial
#################################################
#STRIDE
def STRIDE_naive(scdata,celltype_keys,n_spots,min_cells=2,max_cells=10):
    '''
    params:
    scdata: AnnData
    celltype_keys: .obs['celltype_keys']
    n_spots: n*n
    '''
    n=np.sqrt(n_spots).astype(int)
    x1,x2,spatial = generat_grid2([n,n])
    n_cells = scdata.X.shape[0]
    celltypes = scdata.obs[celltype_keys]
    # 将每个细胞的细胞类型转为onehot
    n_celltypes = celltypes.unique().shape[0]
    onehot_encoder = OneHotEncoder(sparse=True)
    integer_encoded = np.array(celltypes).reshape(len(celltypes), 1)
    cell2celltype = onehot_encoder.fit_transform(integer_encoded)

    #############
    # 每个spot的包含的细胞
    n_spots_cell = np.random.randint(min_cells,max_cells,n_spots)
    indptr = np.concatenate(([0],np.cumsum(n_spots_cell)))
    indices = np.random.choice(n_cells,indptr[-1], replace=True)  #改成True
    elements = np.ones(indices.shape[0])
    # 矩阵中第i行非零元素的列号为indices[indptr[i]:indptr[i+1]]，相应的值为elements[indptr[i]:indptr[i+1]]
    spots2cell = csr_matrix((elements, indices, indptr), shape=(n_spots, n_cells))
    # 比例矩阵
    W = spots2cell @ cell2celltype
    # 每个spot的表达矩阵
    exp_spots2gene = spots2cell @ scdata.X #X 改成 scdata.X
    return W, exp_spots2gene,spatial
########################################################
#stereoscope
def _assemble_spot(cnt : np.ndarray,
                  labels : np.ndarray,
                  alpha : float = 1.0,
                  fraction : float = 1.0,  #修改处
                  bounds : List[int] = [10,30],
                  )->Dict[str,t.Tensor]:

    """Assemble single spot

    generates one synthetic ST-spot
    from provided single cell data

    Parameter:
    ---------
    cnt : np.ndarray
        single cell count data [n_cells x n_genes]
    labels : np.ndarray
        single cell annotations [n_cells]
    alpha : float
        dirichlet distribution
        concentration value
    fraction : float
        fraction of transcripts from each cell
        being observed in ST-spot

    Returns:
    -------
    Dictionary with expression data,
    proportion values and number of
    cells from each type at every
    spot

    """

    # sample between 10 to 30 cells to be present
    # at spot
    n_cells = dists.uniform.Uniform(low = bounds[0],
                                    high = bounds[1]).sample().round().type(t.int)
    n_cells = 1   ####修改
    # get unique labels found in single cell data
    uni_labs, uni_counts = np.unique(labels,
                                     return_counts = True)

    # make sure sufficient number
    # of cells are present within
    # all cell types
    assert np.all(uni_counts >=  30), \
            "Insufficient number of cells"

    # get number of different
    # cell types present
    n_labels = uni_labs.shape[0]

    # sample number of types to
    # be present at current spot
    n_types = dists.uniform.Uniform(low = 1,
                                    high =  n_labels).sample()

    n_types = n_types.round().type(t.int)
    n_types = 1    #####修改
    # select which types to include
    pick_types = t.randperm(n_labels)[0:n_types]
    # pick at least one cell for spot
    members = t.zeros(n_labels).type(t.float)
    while members.sum() < 1:
        # draw proportion values from probability simplex
        member_props = dists.Dirichlet(concentration = alpha * t.ones(n_types)).sample()
        # get integer number of cells based on proportions
        members[pick_types] = (n_cells * member_props).round()

    # get proportion of each type
    props = members / members.sum()
    # convert to ints
    members = members.type(t.int)
    # get number of cells from each cell type

    # generate spot expression data
    spot_expr = t.zeros(cnt.shape[1]).type(t.float32)

    for z in range(n_types):
        # get indices of selected type
        idx = np.where(labels == uni_labs[pick_types[z]])[0]
        # pick random cells from type
        np.random.shuffle(idx)
        idx = idx[0:members[pick_types[z]]]
        # add fraction of transcripts to spot expression
        spot_expr +=  t.tensor((cnt[idx,:]*fraction).sum(axis = 0).round().astype(np.float32))


    return {'expr':spot_expr,
            'proportions':props,
            'members': members,
           }

def assemble_data_set(cnt : pd.DataFrame,
                      labels : pd.DataFrame,
                      n_spots : int,
                      n_genes : int,
                      n_cell_range : List[int],
                      assemble_fun : Callable = _assemble_spot,
                     )-> Dict[str,pd.DataFrame]:

    """Assemble Synthetic ST Data Set

    Assemble synthetic ST Data Set from
    a provided single cell data set

    Parameters:
    ----------
    cnt : pd.DataFrame
        single cell count data
    labels : pd.DataFrame
        single cell annotations
    n_spots : int
        number of spots to generate
    n_genes : int
        number of gens to include
    assemble_fun : Callable
        function to assemble single spot

    """

    # get labels
    ##labels = labels.loc[:,'bio_celltype']
    labels = labels.loc[:,labels.columns[0]]
    # make sure number of genes does not
    # exceed number of genes present
    n_genes = np.min((cnt.shape[1],n_genes))
    # select top expressed genes
    keep_genes = np.argsort(cnt.sum(axis=0))[::-1]
    keep_genes = keep_genes[0:n_genes]
    cnt = cnt.iloc[:,keep_genes]

    # get unique labels
    uni_labels = np.unique(labels.values)
    n_labels = uni_labels.shape[0]

    # prepare matrices
    st_cnt = np.zeros((n_spots,cnt.shape[1]))
    st_prop = np.zeros((n_spots,n_labels))
    st_memb = np.zeros((n_spots,n_labels))

    #np.random.seed(1337)
    #t.manual_seed(1337)
    # generate one spot at a time
    for spot in range(n_spots):
        spot_data = assemble_fun(cnt.values,
                                 labels.values,
                                 bounds = n_cell_range,
                                 )

        st_cnt[spot,:] = spot_data['expr']
        st_prop[spot,:] = spot_data['proportions']
        st_memb[spot,:] =  spot_data['members']

        index = pd.Index(['Spotx' + str(x + 1) for \
                          x in range(n_spots) ])

    # convert to pandas DataFrames
    st_cnt = pd.DataFrame(st_cnt,
                          index = index,
                          columns = cnt.columns,
                         )

    st_prop = pd.DataFrame(st_prop,
                           index = index,
                           columns = uni_labels,
                          )
    st_memb = pd.DataFrame(st_memb,
                           index = index,
                           columns = uni_labels,
                           )


    return {'counts':st_cnt,
            'proportions':st_prop,
            'members':st_memb}


def stereoscope_naive(sc_cnt_pth = None, sc_lbl_pth = None, out_dir = None, data = None):
    n_spots = len(data)
    n_genes = data.X.shape[1]
    # read data
    sc_cnt = pd.read_csv(sc_cnt_pth,
                         #sep = '\t',
                         index_col = 0,
                         header = 0)

    sc_lbl = pd.read_csv(sc_lbl_pth,
                         #sep = '\t',
                         index_col = 0,
                         header = 0)
    # match count and label data
    inter = sc_lbl.index
    labels = sc_lbl.iloc[:,0].values
    # get unique labels
    uni_labs, uni_counts = np.unique(labels, return_counts = True)
    # only keep types with more than 2*30 cells
    keep_types = uni_counts > 30*2
    keep_cells = np.isin(labels, uni_labs[keep_types])
    labels = labels[keep_cells]
    sc_cnt = sc_cnt.iloc[keep_cells,:]
    sc_lbl = sc_lbl.iloc[keep_cells,:]
    uni_labs, uni_counts = np.unique(labels, return_counts = True)
    n_types = uni_labs.shape[0]
    assert np.all(uni_counts > 2), \
                "Only one cell in types"

    # get member indices for each set
    idx_generation = []
    idx_validation = []
    #np.random.seed(1337)
    for z in range(n_types):
        tmp_idx = np.where(labels == uni_labs[z])[0]
        n_generation = int(round(tmp_idx.shape[0] / 2 ))

        idx_generation += tmp_idx[0:n_generation].tolist()
        idx_validation += tmp_idx[n_generation::].tolist()
    idx_generation.sort()
    idx_validation.sort()
    # make sure no members overlap between sets
    assert len(set(idx_generation).intersection(set(idx_validation))) == 0, \
            "validation and genreation set are not orthogonal"
    # assemble sets from indices
    cnt_validation = sc_cnt.iloc[idx_validation,:]
    cnt_generation = sc_cnt.iloc[idx_generation,:]

    lbl_validation = sc_lbl.iloc[idx_validation,:]
    lbl_generation = sc_lbl.iloc[idx_generation,:]

    # save sets
    cnt_validation.to_csv(osp.join(out_dir,'.'.join(['validation',osp.basename(sc_cnt_pth)])),
                          #sep = '\t',
                          header = True,
                          index = True,
                          index_label = 'cell',)

    cnt_generation.to_csv(osp.join(out_dir,'.'.join(['generation',osp.basename(sc_cnt_pth)])),
                          #sep = '\t',
                          header = True,
                          index = True,
                          index_label = 'cell')

    lbl_validation.to_csv(osp.join(out_dir,'.'.join(['validation',osp.basename(sc_lbl_pth)])),
                          #sep = '\t',
                          header = True,
                          index = True,
                          index_label = 'cell')

    lbl_generation.to_csv(osp.join(out_dir,'.'.join(['generation',osp.basename(sc_lbl_pth)])),
                          #sep = '\t',
                          header = True,
                          index = True,
                          index_label = 'cell')
    
    sc_cnt = pd.read_csv(sc_cnt_pth,
                     #sep = '\t',
                     index_col = 0,
                     header = 0)

    sc_lbl = pd.read_csv(sc_lbl_pth,
                     #sep = '\t',
                     index_col = 0,
                     header = 0)
    assembled_set = assemble_data_set(sc_cnt,
                                  sc_lbl,
                                  n_spots = n_spots,
                                  n_genes = n_genes,
                                  n_cell_range = [0,1],
                                  assemble_fun = _assemble_spot,
                                  )
    for k,v in assembled_set.items() :
        out_pth = osp.join(out_dir, '.'.join([k,'st_synth','tsv']))
        v.to_csv(out_pth,
                    #sep = '\t',
                    index = True,
                    header = True,
                   )
    return None		



#import squidpy as sq
import scanpy as sc
import anndata as ad
import numpy as np
import pandas as pd
import squidpy as sq
import matplotlib.pyplot as plt
import scipy
import random 
import matplotlib
import seaborn as sns
import math
#import time
from scipy.special import softmax
from scipy.optimize import minimize
from numpy.random import uniform
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import LabelEncoder


import numba
import warnings
warnings.filterwarnings("ignore", message=".*The 'nopython' keyword.*")
def get_ct_sample(Num_celltype = None, Num_sample = None, prior = None):
    if prior is None:
        prior = np.ones(Num_celltype)/Num_celltype
    else:
        assert len(prior) == Num_celltype, "please check your prior length, that must be equal to the number of celltype"
        assert sum(prior)<= 1, "Sum of prior is out of range!"
        assert 1-sum(prior) <1e-3,"Sum of prior is equal to 1"
    
    Num_ct_sample = np.zeros(Num_celltype,dtype = np.int32)
    Num_ct_sample[0:(Num_celltype-1)] = (prior[0:(Num_celltype-1)]*Num_sample+0.5*np.ones(Num_celltype-1)).astype(int)
    Num_ct_sample[Num_celltype-1] = Num_sample-np.sum(Num_ct_sample[0:(Num_celltype-1)])
    return Num_ct_sample

def addictive_freq(n_c):
    target_freq = np.ones((n_c,n_c))
    for i in np.arange(n_c):
        target_freq[i,i] = 4*(n_c-1)
    
    return target_freq/np.sum(target_freq,axis=1,keepdims=True)

def exclusive_freq(n_c):
    target_freq = np.ones((n_c,n_c))
    for i in np.arange(n_c):
        target_freq[i,i] = 3*(n_c-1)
        if i%2 == 1:
            target_freq[i-1,i] = 3*(n_c-1)
            target_freq[i,i-1] = 3*(n_c-1)
    
    return target_freq/np.sum(target_freq,axis=1,keepdims=True)

def stripe_freq(n_c):
    target_freq = np.ones((n_c,n_c))
    for i in np.arange(n_c):
        target_freq[i,i] = 3*(n_c-1)
        if i>0:
            target_freq[i-1,i] = 3*(n_c-1)
    
    return target_freq/np.sum(target_freq,axis=1,keepdims=True)

## 把squidpy的参数全部拿过来

def get_spaital_network(Num_sample = None, spatial = None,n_neighs = None,radius = None,coord_type = "grid",n_rings = 2,set_diag = False):
    spatial_adata = ad.AnnData(np.empty((Num_sample,1),dtype="float32"))
    spatial_adata.obsm["spatial"] = spatial
    sq.gr.spatial_neighbors(spatial_adata, n_rings=n_rings, coord_type=coord_type, n_neighs=n_neighs, radius=radius,set_diag =set_diag)
    sn = spatial_adata.obsp["spatial_connectivities"]
    
    return sn

def init_ct(Num_celltype = None,Num_ct_sample = None, seed = None):
    if seed is not None:
        np.random.seed(seed)
    init_assign = np.repeat(np.arange(Num_celltype), Num_ct_sample)
    np.random.shuffle(init_assign)
    return init_assign

def get_onehot_ct(init_assign = None):
    label_encoder = LabelEncoder()
    integer_encoded = label_encoder.fit_transform(init_assign)
    onehot_encoder = OneHotEncoder(sparse=False)
    integer_encoded = integer_encoded.reshape(len(integer_encoded), 1)
    onehot_ct = onehot_encoder.fit_transform(integer_encoded)
    return onehot_ct.astype(np.float32)

def get_trans(adata = None, ct = None):
    sn = get_spaital_network(Num_sample=adata.obs.shape[0],
                         spatial=adata.obsm["spatial"], coord_type = "generic",
                         n_neighs=8)
    onehot_ct = get_onehot_ct(init_assign=ct)
    nb_count = np.array(sn * onehot_ct, dtype=np.float32)
    target_trans = get_nb_freq(nb_count=nb_count, onehot_ct=onehot_ct)
    return target_trans

@numba.jit("float32[:, ::1](float32[:, ::1], float32[:, ::1])")
def get_nb_freq( nb_count = None, onehot_ct = None):
#     nb_freq = onehot_ct.T @ nb_count
    nb_freq = np.dot(onehot_ct.T, nb_count)
    res = nb_freq/nb_freq.sum(axis = 1).reshape(onehot_ct.shape[1],-1)
    return res


def swap_ct(celltype_assignment = None, Num_celltype = None,swap_num = None):
    swap_cluster = np.random.choice(np.arange(Num_celltype),2,replace = False)
    swap_i_index = np.random.choice(np.where(celltype_assignment==swap_cluster[0])[0],swap_num,replace = False)
    swap_j_index = np.random.choice(np.where(celltype_assignment==swap_cluster[1])[0],swap_num,replace = False)
    
    return (swap_i_index,swap_cluster[0]), (swap_j_index,swap_cluster[1])
    
def get_swap_nb_count(  nb_count = None,swap_i = None, swap_j = None,sn = None):
    
    swap_i_nb_index = sn[swap_i[0]].indices
    nb_count[swap_i_nb_index,swap_i[1]] -= 1
    nb_count[swap_i_nb_index,swap_j[1]] += 1
    
    swap_j_nb_index = sn[swap_j[0]].indices
    nb_count[swap_j_nb_index,swap_j[1]] -= 1
    nb_count[swap_j_nb_index,swap_i[1]] += 1
    
    return nb_count






#import squidpy as sq
import scanpy as sc
import anndata as ad
import numpy as np
import pandas as pd
import squidpy as sq
import matplotlib.pyplot as plt
import scipy
import random 
import matplotlib
import seaborn as sns
import math
#import time
from scipy.special import softmax
from scipy.optimize import minimize
from numpy.random import uniform
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import LabelEncoder
from .utils import *
from .Annealing import *
from .enhance import *
from scipy.sparse import coo_matrix,csr_matrix


def get_mesh_counts(locations = None, grid_row = None, grid_col = None,
                   image_width = None, image_height = None):
    
    grid_num = grid_row *grid_col
    
    if image_width is None and image_height is None:
        
        image_width  = np.max(locations[:,0]).round().astype(int)
        image_height = np.max(locations[:,1]).round().astype(int)
 
    dx = image_width / grid_row  
    dy = image_height / grid_col
    points_x_indices = (locations[:,0] / dx).astype(int)
    points_y_indices = (locations[:,1] / dy).astype(int)
    
#     print("cell location sort by meshgrid")
#     points_sortidex = np.lexsort([points_y_indices,points_x_indices])
   
#     res = csr_matrix((np.ones_like(points_x_indices, dtype=np.int), 
#                         (points_x_indices,points_y_indices )), 
#                         shape=(grid_row ,grid_col)).A
#     counts = res.reshape(-1)

    ## 细胞的spot编号
    cell_spot_id = (points_x_indices*grid_col + points_y_indices).astype(int)
    
    ## 细胞的编号
    cell_id = np.arange(cell_spot_id.size)
    
    ##  替换np.where 每一个grid  obtain 每一个spot的cell的index 行cell_id 列spot 
    cell_spot_idx_matrix = csr_matrix(
        (np.ones(cell_spot_id.size), (cell_id,cell_spot_id)),
        shape=(cell_spot_id.size,grid_num))
    
   
    a, b = np.meshgrid(np.arange(grid_row),
                   np.arange(grid_col),
                   indexing='ij')
    
    grid_row_loc = a.reshape(-1) * dx
    grid_col_loc = b.reshape(-1) * dy
    
    return  cell_spot_idx_matrix, grid_row_loc, grid_col_loc, image_width, image_height


def simulate_10X(cell_num=None,
                 Num_celltype=None,
                 prior=None,
                 target_trans=None,
                 spot_radius=None,
                 spot_min=None,
                 spot_max=None,
                 image_width=None,
                 image_height=None,
                 cell_location=None,
                 tol=2e-2,
                 T=1e-2,
                 loop_times=None,
                 smallsample_max_iter=80000,
                 bigsample_max_iter=10000,
                 ref=None):
    '''
    Problems: 1. this function, you must give `image_width` and `image_height`, but 
                 function `get_mesh_counts` write the situation that these paras are given
                 by yourself.
              2. How to control `spot_min` and `spot_max`
    '''

    # we will compute meshgrid nums to determine how many
    ## 用户从HE学出细胞位置，通过cell_location给定
    ## 也可以基于细胞数 随机设定
    ## 判断 cell_num是否为空，补
    ## 判断cell_num与cell_location同时存在,num = cell_location.shape[0]
    if ref is not None:
        cell_num = ref.shape[0]
        cell_location = ref.obsm["spatial"]
        Num_ct_sample = np.bincount(ref.obs.label)
        randcelltype = init_ct(Num_celltype=len(Num_ct_sample),
                               Num_ct_sample=Num_ct_sample)
        Num_celltype = len(Num_ct_sample)
        if cell_num <= 10000:
            loop_times = 1
        else:
            grid_row = np.sqrt(cell_num).astype(int)
            grid_col = grid_row
            grid_num = grid_row * grid_col
            cell_spot_idx_matrix, grid_row_loc, grid_col_loc, image_width, image_height = get_mesh_counts(
                locations=cell_location,
                grid_row=grid_row,
                grid_col=grid_col,
                image_width=image_width,
                image_height=image_height)

            # cell_location = cell_location[points_sortidex, :]

            print("enhance resolution based on meshgrid")
            ## 注意此时grid_row *grid_col != grid_spatial
            grid_spatial = np.vstack((grid_row_loc, grid_col_loc)).T

            celltype_assignment, _, loop_times = enhance_loop(
                Num_sample=grid_num,
                Num_celltype=Num_celltype,
                prior=prior,
                target_trans=target_trans,
                original_grid=grid_spatial,
                grid_row=grid_row,
                grid_col=grid_col,
                loop_times=None,
                windows_row_list=None,
                windows_col_list=None,
                swap_num_list=None,
                tol_list=None,
                T=1e-5,
                smallsample_max_iter=smallsample_max_iter,
                bigsample_max_iter=bigsample_max_iter)

            print(
                "cell type labels of the grid are inversely assigned to cells, and adjust ratio."
            )

            ## 得到网格的细胞类型的onehot
            grid_ct_onehot = get_onehot_ct(init_assign=celltype_assignment.reshape(-1))
            randcelltype = np.argmax(cell_spot_idx_matrix * grid_ct_onehot, axis=1)
            # randcelltype = np.repeat(celltype_assignment, counts)

            ##  adjust ratio
            # Num_ct_sample = get_ct_sample(Num_celltype=Num_celltype,
            #                               Num_sample=cell_num,
            #                               prior=prior)
            randcelltype = mutate(celltype_assignment=randcelltype,
                                  Num_ct_sample=Num_ct_sample)

    else:
        if cell_location is not None:
            cell_num = cell_location.shape[0]
        else:
            points_x = np.random.uniform(0, image_width, cell_num)
            points_y = np.random.uniform(0, image_height, cell_num)
            cell_location = np.vstack((points_x, points_y)).T
        if cell_num <= 10000:

            ##  adjust ratio
            Num_ct_sample = get_ct_sample(Num_celltype=Num_celltype,
                                          Num_sample=cell_num,
                                          prior=prior)
            randcelltype = init_ct(Num_celltype = Num_celltype,
                                   Num_ct_sample = Num_ct_sample )

            loop_times = 1

        else:
            grid_row = np.sqrt(cell_num).astype(int)
            grid_col = grid_row
            grid_num = grid_row * grid_col
            cell_spot_idx_matrix, grid_row_loc, grid_col_loc, image_width, image_height  = get_mesh_counts(
                locations=cell_location,
                grid_row=grid_row,
                grid_col=grid_col,
                image_width=image_width,
                image_height=image_height)

           # cell_location = cell_location[points_sortidex, :]

            print("enhance resolution based on meshgrid")
            ## 注意此时grid_row *grid_col != grid_spatial
            grid_spatial = np.vstack((grid_row_loc, grid_col_loc)).T

            celltype_assignment, _, loop_times = enhance_loop(
                Num_sample=grid_num,
                Num_celltype=Num_celltype,
                prior=prior,
                target_trans=target_trans,
                original_grid=grid_spatial,
                grid_row=grid_row,
                grid_col=grid_col,
                loop_times=None,
                windows_row_list=None,
                windows_col_list=None,
                swap_num_list=None,
                tol_list=None,
                T=1e-5,
                smallsample_max_iter=smallsample_max_iter,
                bigsample_max_iter=bigsample_max_iter)

            print(
                "cell type labels of the grid are inversely assigned to cells, and adjust ratio."
            )

            ## 得到网格的细胞类型的onehot
            grid_ct_onehot = get_onehot_ct(init_assign = celltype_assignment.reshape(-1) )
            randcelltype = np.argmax(cell_spot_idx_matrix * grid_ct_onehot,axis = 1)
            #randcelltype = np.repeat(celltype_assignment, counts)

            ##  adjust ratio
            Num_ct_sample = get_ct_sample(Num_celltype=Num_celltype,
                                          Num_sample=cell_num,
                                          prior=prior)
            randcelltype = mutate(celltype_assignment=randcelltype,
                                  Num_ct_sample=Num_ct_sample)

    randsn = get_spaital_network(Num_sample=cell_num,
                                 spatial=cell_location, coord_type = "generic",
                                 n_neighs=8)

    onehot_ct = get_onehot_ct(init_assign=randcelltype)
    nb_count = np.array(randsn * onehot_ct, dtype=np.float32)

    swap_num = 2 ** (loop_times-1)
    print("swap_num: ", swap_num)
    celltype_assignment = STsim(Num_sample=cell_num,
                             Num_celltype=Num_celltype,
                             celltype_assignment=randcelltype,
                             target_trans=target_trans,
                             T=T,
                             chain_len=100,
                             error=1e2,
                             onehot_ct=onehot_ct,
                             tol=tol,
                             decay=0.5,
                             nb_count=nb_count,
                             sn=randsn,
                             swap_num=swap_num,
                             smallsample_max_iter=smallsample_max_iter,
                             bigsample_max_iter=bigsample_max_iter)
    

    return celltype_assignment, cell_location


def extract_loc(img = None,
                layer = None,
                library_id = None,
                channel =  0,
                ):
    '''
    Extract cell locations from an image. 
    
    用squidpy做完细胞分割后，利用该函数可以实现获取细胞像素坐标，之前的图像处理流程参考:
    https://squidpy.readthedocs.io/en/latest/auto_examples/image/compute_segment_hne.html
    Parameters
    ----------
    %(img_container)s
    %(img_layer)s
    layer
        Image layer in ``img`` that should be processed. If `None` and only 1 layer is present, it will be selected.
    %(library_id)s
        If `None`, all Z-dimensions are processed at once, treating the image as a 3D volume.
    channel
        Channel index to use for extracting. If `None`, use all channels.
    Returns
    -------
    Return an array with cell locations extracted from image. the frist column represents the
    y_coord of cell in the image, the second column represents the x_coord of cell in the image.
    '''
    layer = img._get_layer(layer)
    library_id = img._get_library_ids(library_id)
    if channel is None and img[layer].shape[-1] > 1:
        raise ValueError("Watershed segmentation does not work with multiple channels.")
        
    segments_matrix = np.array(img["segmented_watershed"])
    segments_matrix =  coo_matrix(segments_matrix.reshape(segments_matrix.shape[0:2]))
    segments_matrix = pd.DataFrame({"pixel_y":segments_matrix.row,
                             "pixel_x":segments_matrix.col,
                              "segment_label":segments_matrix.data})
    cell_location = segments_matrix.groupby(segments_matrix["segment_label"]).mean().values 
    print(f"Number of cells in image: {cell_location.shape[0]}")
    return cell_location


import warnings
import sys
import os
#from .joint_simulator import Simulator, get_gene_prior, get_nf_prior
#from .opt import valid_neighbourhood_frequency
from os.path import join
import time
from .sim_expr import *
from .random_based_utils import *

warnings.filterwarnings("ignore")
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))


def get_sim_spot_level_expr( spot_num = None, spot_row = None, spot_col = None,
                            Num_sample = None, spot_diameter = None, image_width = None,
                            image_height = None,celltype_assignment = None,coord_type = None,
                            cell_spatial = None,sim_cell_expr = None,gap = 0,
                           spot_generate_type = "square",cell_coord_type = "grid",
                            custom_spot_loc=None, grid_row = None, grid_col = None):
    if spot_generate_type == "square" and cell_coord_type == "generic":
        ## 随机点的网格划分，不是网格merge 

        grid_row = np.ceil(image_width / spot_diameter).astype(int)
        grid_col = np.ceil(image_height / spot_diameter).astype(int)
        grid_num = grid_row * grid_col
        points_x_indices = (cell_spatial[:, 0] / spot_diameter).astype(int)
        points_y_indices = (cell_spatial[:, 1] / spot_diameter).astype(int)

        ## 细胞的spot编号
        cell_spot_id = (points_x_indices*grid_col + points_y_indices).astype(int)
        ## 细胞的编号
        cell_id = np.arange(cell_spot_id.size)


        ##  替换np.where 每一个grid  obtain 每一个spot的cell的index 行spot 列cell_id
        spot_cell_idx_matrix = csr_matrix(
            (np.ones(cell_spot_id.size), (cell_spot_id, cell_id)),
            shape=(grid_num, cell_spot_id.size))

        spot_expr = spot_cell_idx_matrix * sim_cell_expr.X

        onehot_ct = get_onehot_ct(init_assign=celltype_assignment)

        ## 注意有些spot里面可能没有cell
        spot_ct_count = spot_cell_idx_matrix * onehot_ct

        ## 取网格的中心点作为spot坐标
        a, b = np.meshgrid(np.arange(grid_row),
                       np.arange(grid_col),
                       indexing='ij')
        spot_pixel_row = (a.reshape(-1)+0.5)*spot_diameter
        spot_pixel_col = (b.reshape(-1)+0.5)*spot_diameter

        spot_loc =  np.vstack((spot_pixel_row, spot_pixel_col)).T

        return spot_expr, spot_loc, spot_cell_idx_matrix, spot_ct_count

    if spot_generate_type == "circle" and cell_coord_type == "generic":

     ####### 规定直径 and gap

        grid_row = int(image_width / gap)
        grid_col = int(image_height / gap)
        grid_num = grid_row * grid_col

        ## 取网格的中心点作为spot坐标
        a, b = np.meshgrid(np.arange(grid_row),
                       np.arange(grid_col),
                       indexing='ij')
        spot_pixel_row = (a.reshape(-1)+0.5)*gap
        spot_pixel_col = (b.reshape(-1)+0.5)*gap

        spot_loc =  np.vstack((spot_pixel_row, spot_pixel_col)).T

        cellandspot_loc =  np.vstack((cell_spatial,spot_loc))

        ## get中心点的邻居
        sn = get_spaital_network(Num_sample=cellandspot_loc.shape[0],spatial=cellandspot_loc,
                         coord_type = "generic",n_rings = 1,set_diag=False,
                                 radius = spot_diameter/2)

        spot_cell_idx_matrix = sn[Num_sample:,Num_sample]

        spot_expr = spot_cell_idx_matrix * sim_cell_expr.X

        onehot_ct = get_onehot_ct(init_assign=celltype_assignment)

        ## 注意有些spot里面可能没有cell
        spot_ct_count = spot_cell_idx_matrix * onehot_ct

        return spot_expr, spot_loc, spot_cell_idx_matrix

    if spot_generate_type == "square" and cell_coord_type == "grid":

      ##### 还有网格merge
        if grid_row is None and grid_col is None:
            grid_row = np.sqrt(Num_sample).astype(int)
            grid_col = grid_row
        ## 网格merge,我们取九个格子merge 类似ST 同时设置spot之间的gap
        windows_row = 3
        windows_col = 3
        gap = gap
        spot_diameter = 1

        ## 边缘没有3*3的网格,设不设置gap?
        spot_row_num = int(grid_row/(windows_row+gap))
        spot_col_num = int(grid_col/(windows_col+gap))

        a, b = np.meshgrid(np.arange(grid_row),
                   np.arange(grid_col),
                   indexing='ij')
        ## lra 与lrb是9个网格的中心网格 实际是中心网格左下角的点的坐标
        lra = a[np.arange(1, spot_row_num*(windows_row+gap), (windows_row+gap)
                  ), :len(np.arange(1, spot_col_num*(windows_col+gap), (windows_col+gap)))].reshape(-1)
        lrb = b[:len(np.arange(1, spot_row_num*(windows_row+gap), (windows_row+gap))),
            np.arange(1, spot_col_num*(windows_col+gap), (windows_col+gap))].reshape(-1)

        ## 中心网格标号
        lrspot_layer_idx = lra*grid_col+lrb

        ## get中心点的邻居，即9个网格 包括自己，
        sn = get_spaital_network(Num_sample=cell_spatial.shape[0],spatial=cell_spatial,
                         n_neighs=8,coord_type = "grid",n_rings = 1,set_diag=True)

        spot_cell_idx_matrix = sn[lrspot_layer_idx]
        spot_expr = spot_cell_idx_matrix * sim_cell_expr.X

        onehot_ct = get_onehot_ct(init_assign=celltype_assignment)

        ## 注意有些spot里面可能没有cell
        spot_ct_count = spot_cell_idx_matrix * onehot_ct

        spot_pixel_row = (lra.reshape(-1)+0.5)*spot_diameter
        spot_pixel_col = (lrb.reshape(-1)+0.5)*spot_diameter
        spot_loc = np.vstack((spot_pixel_row, spot_pixel_col)).T


        return spot_expr, spot_loc, spot_cell_idx_matrix

    if custom_spot_loc is not None:
        ## 以及指定spot坐标set  

        ## must numpy shape(n*2)
        ## custom_spot_loc and spot_diameter
        spot_loc = custom_spot_loc

        cellandspot_loc =  np.vstack((cell_spatial,spot_loc ))
        sn = get_spaital_network(Num_sample=cellandspot_loc.shape[0],spatial=cellandspot_loc,
                         coord_type = "generic",n_rings = 1,set_diag=False,
                                 radius = spot_diameter/2)

        spot_cell_idx_matrix = sn[cell_spatial.shape[0]:,cell_spatial.shape[0]]

        spot_expr = spot_cell_idx_matrix * sim_cell_expr.X

        onehot_ct = get_onehot_ct(init_assign=celltype_assignment)

        ## 注意有些spot里面可能没有cell
        spot_ct_count = spot_cell_idx_matrix * onehot_ct

    return spot_expr, spot_loc, spot_cell_idx_matrix


# def rctd(adata=None, ctkey=None, method=None, file_path=None, seed=123, ):
#     W, exp_spots2gene, spatial, spots2cell = RCTD_naive(scdata=adata,
#                                                         row=row_col_size,
#                                                         col=row_col_size,
#                                                         Min=0,
#                                                         Max=1,
#                                                         ctkey=ctkey,
#                                                         maxctnum=len(adata.obs.celltype.unique()))
#     exp_spots2gene = downsample_matrix_by_cell(exp_spots2gene, 10000)
#     simulatedat = sc.AnnData(exp_spots2gene[:len(adata), :])
#     simulatedat.obs = pd.DataFrame(
#         np.array(W[:len(adata), :] @ range(len(adata.obs.celltype.unique())), dtype=int))


def sim_naive_cell(use_real_adata=None,
              ctkey=None,method=None,
              file_path=None,seed = 123):
    global simulatedat
    np.random.seed(seed)
    out_dir = file_path+'newsimu/'
    row_col_size = round(np.sqrt(len(use_real_adata)))+1
    #choose method
    if method == 'RCTD':
        W,exp_spots2gene,spatial,spots2cell = RCTD_naive(scdata=use_real_adata,
                                                     row=row_col_size,
                                                     col=row_col_size,
                                                     Min = 0,
                                                     Max = 1,
                                                     ctkey=ctkey,
                                                     maxctnum = len(use_real_adata.obs.celltype.unique()))
        exp_spots2gene = downsample_matrix_by_cell(exp_spots2gene,10000)
        simulatedat = sc.AnnData(exp_spots2gene[:len(use_real_adata),:])
        simulatedat.obs = pd.DataFrame(np.array(W[:len(use_real_adata),:] @ range(len(use_real_adata.obs.celltype.unique())),dtype = int))

    if method == 'STRIDE':
        W,exp_spots2gene,spatial = STRIDE_naive(use_real_adata, ctkey, row_col_size*row_col_size, min_cells=1, max_cells=2)
        exp_spots2gene = downsample_matrix_by_cell(exp_spots2gene,10000)
        simulatedat = sc.AnnData(exp_spots2gene[:len(use_real_adata),:])
        simulatedat.obs = pd.DataFrame(np.array(W[:len(use_real_adata),:] @ range(len(use_real_adata.obs.celltype.unique())),dtype = int))

    if method == 'sterepscope':
        ct = use_real_adata.obs.celltype.unique()
        use_exp = pd.DataFrame(use_real_adata.X,columns=use_real_adata.var.gene)
        use_exp.to_csv(file_path+str(len(ct))+'ct_real_exp.csv')
        use_label = use_real_adata.obs.celltype.to_frame()
        use_label.to_csv(file_path+str(len(ct))+'ct_real_celltype.csv')
        sc_cnt_pth = file_path+str(len(use_real_adata.obs.celltype.unique()))+'ct_real_exp.csv'
        sc_lbl_pth = file_path+str(len(use_real_adata.obs.celltype.unique()))+'ct_real_celltype.csv'
        stereoscope_naive(sc_cnt_pth = sc_cnt_pth,
                      sc_lbl_pth = sc_lbl_pth,
                      out_dir = out_dir,
                      data = use_real_adata)
        data1 = pd.read_csv(out_dir+'counts.st_synth.tsv',index_col = 0)
        data1 = data1.sort_index(axis=1)
        menber = pd.read_csv(out_dir+'members.st_synth.tsv',index_col = 0)
        simulatedat = sc.AnnData(data1)
        simulatedat.obs = pd.DataFrame(np.array(menber @ range(len(use_real_adata.obs.celltype.unique())),dtype = int))

    #save
    if np.isin(method,['RCTD','STRIDE','stereoscope']):
        simulatedat.obs.columns = ['label']
        simulatedat.obs.label = simulatedat.obs.label.astype(int).astype('category')
        simulatedat.obs['celltype'] = use_real_adata.obs[ctkey]
        for j in range(len(use_real_adata.obs[ctkey].unique())):
            simulatedat.obs['celltype'][np.where(simulatedat.obs['label'] == j)[0]] = np.unique(use_real_adata.obs[ctkey])[j]
        simulatedat.obs.index = ['cell'+str(i+1) for i in range(use_real_adata.shape[0])]

        simulatedat.obsm = use_real_adata.obsm  #real location
        simulatedat.uns['W'] = W[:len(use_real_adata),:]
        simulatedat.uns['celltype_name'] = np.unique(use_real_adata.obs[ctkey])
        simulatedat.var = use_real_adata.var
        # if not os.path.exists(file_path+'newsimu/'):
        #     os.mkdir(file_path+'newsimu/')
        simulatedat.write(file_path+method+'_simu_cell_level'+'.h5ad')

    return simulatedat

def sim_naive_spot(use_real_adata=None,level='cell',spot_diameter=500,
                   ctkey=None,method=None,file_path=None,seed = 123):
    global sim_spot_adata
    sim_adata = sim_naive_cell(use_real_adata,ctkey,method,file_path,seed)

    if level == "cell":
        return sim_adata
    elif level == 'spot':
        ## transform coordinate
        temp1 = np.min(sim_adata.obsm["spatial"][:,0])
        temp2 = np.min(sim_adata.obsm["spatial"][:,1])
        sim_adata.obsm["spatial"][:,0] = sim_adata.obsm["spatial"][:,0]- temp1
        sim_adata.obsm["spatial"][:,1] = sim_adata.obsm["spatial"][:,1]- temp2
        spot_expr, spot_loc, spot_cell_idx_matrix,spot_ct_count  = get_sim_spot_level_expr(spot_num=None,
                    spot_row=None,
                    spot_col=None,
                    Num_sample=sim_adata.shape[0],
                    spot_diameter=spot_diameter,
                    image_width=np.max(sim_adata.obsm["spatial"][:,0]),
                    image_height=np.max(sim_adata.obsm["spatial"][:,1]),
                    celltype_assignment=sim_adata.obs.label,
                    coord_type="generic",
                    cell_spatial=sim_adata.obsm["spatial"],
                    sim_cell_expr=sim_adata,
                    gap=0,
                    spot_generate_type="square",
                    cell_coord_type="generic",
                    grid_row=None,
                    grid_col=None)
        mask = np.all(spot_ct_count == 0, axis=1)
        ## adjust spot_loc
        spot_loc[:,0] = spot_loc[:,0] + temp1
        spot_loc[:,1] = spot_loc[:,1] + temp2
        sim_spot_adata = ad.AnnData(spot_expr)
        sim_spot_adata.obsm["spatial"] = spot_loc
        sim_spot_adata.var = sim_adata.var
        sim_spot_adata.obs.index =  ['spot'+str(j) for j in range(spot_expr.shape[0])]
        sim_spot_adata = sim_spot_adata[~mask]
        sim_spot_adata.uns["W"] = spot_ct_count[~mask]
        sim_spot_adata.write_h5ad(file_path + method +"_simu_spot_level" + ".h5ad")
        return sim_spot_adata
    else:
        raise ValueError("The parameter level must be either 'cell' or 'spot'")

#generate gene expression using splatter
from .scsim import scsim
"""
This part is modified code from 'Identifying gene expression programs of cell-type identity and cellular activity with single-cell RNA-Seq'
from this repository https://github.com/dylkot/scsim
Modified version allows the simulator take a pre-defined cell group assignment.
"""
def sim_naive_spot_splatter(use_real_adata=None,level='cell',spot_diameter=500,
                   ctkey=None,method=None,file_path=None,seed = 123, doubletfrac = 0,
                    deloc = 1, progdeloc = 1, descale = 1.0, progcellfrac = .35,
                    deprob = .025, nproggenes = 400):
    adata_simu = sim_naive_spot(use_real_adata, level, spot_diameter,
                                   ctkey, method, file_path, seed)
    #gen exp using splatter
    ncells = adata_simu.shape[0]
    ngenes = adata_simu.shape[1]
    K = len(adata_simu.obs.celltype.unique())
    groupid = np.array(adata_simu.obs.label.values)
    ndoublets = int(doubletfrac*ncells)
    nproggroups = K
    proggroups = list(range(1, nproggroups+1))
    simulator = scsim(ngenes=ngenes, ncells=ncells, ngroups=K, groupid = groupid, libloc=7.64, libscale=0.78,
                         mean_rate=7.68,mean_shape=0.34, expoutprob=0.00286,
                         expoutloc=6.15, expoutscale=0.49,
                         diffexpprob=deprob, diffexpdownprob=0., diffexploc=deloc, diffexpscale=descale,
                         bcv_dispersion=0.448, bcv_dof=22.087, ndoublets=ndoublets,
                         nproggenes=nproggenes, progdownprob=0., progdeloc=progdeloc,
                         progdescale=descale, progcellfrac=progcellfrac, proggoups=proggroups,
                         minprogusage=.1, maxprogusage=.7, seed=seed)
    start_time = time.time()
    simulator.simulate()
    end_time = time.time()
    print("Elapsing time is %.2f"%(end_time - start_time))
    gene_expression_new = np.asarray(simulator.counts)
    adata_simu.X = gene_expression_new
    return adata_simu


#import squidpy as sq
import scanpy as sc
import anndata as ad
import numpy as np
import pandas as pd
import squidpy as sq
import matplotlib.pyplot as plt
import scipy
import random 
import matplotlib
import seaborn as sns
import math
#import time
from scipy.special import softmax
from scipy.optimize import minimize
from numpy.random import uniform
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import LabelEncoder
from .utils import *

def STsim(Num_sample=None,
          Num_celltype = None,
          celltype_assignment=None,
          target_trans = None,
          T=1000,
          chain_len=100,
          error=1e2,
          tol=2e-2,
          decay=0.5,
          onehot_ct=None,
          nb_count=None,
          sn=None,
          swap_num=None,
          smallsample_max_iter=None,
          bigsample_max_iter=None):
    '''
        smallsample: <10000 and swap_num = 1
        bigsample: >=10000
    '''

    error_record = []
    iter_n = 0
    print("Refine cell type using Metropolis–Hastings algorithm.")
    print("Sample num"+":{}".format(Num_sample) )


    nb_freq = get_nb_freq(nb_count=nb_count, onehot_ct=onehot_ct)
    dist_before = sum(scipy.stats.entropy(target_trans, nb_freq))

    if Num_sample <= 10000:
        swap_num = 1
        max_iter = smallsample_max_iter
        criterion3 = True
        same_errlen = 250
        while error > tol and iter_n < max_iter and criterion3:

            for i in range(chain_len):

                before_ct_assign, before_onehot_ct, before_nb_count = celltype_assignment.copy(), onehot_ct.copy(), nb_count.copy()

                swap_i, swap_j = swap_ct(
                    celltype_assignment=celltype_assignment,
                    Num_celltype=Num_celltype,
                    swap_num=swap_num)

                celltype_assignment[swap_i[0]], celltype_assignment[
                    swap_j[0]] = celltype_assignment[
                        swap_j[0]], celltype_assignment[swap_i[0]]
                onehot_ct[swap_i[0]], onehot_ct[swap_j[0]] = onehot_ct[
                    swap_j[0]], onehot_ct[swap_i[0]]

                nb_count = get_swap_nb_count(nb_count=nb_count,
                                             swap_i=swap_i,
                                             swap_j=swap_j,
                                             sn=sn)
                # nb_count = sn * onehot_ct
                nb_freq = get_nb_freq(nb_count=nb_count, onehot_ct=onehot_ct)

                dist_after = sum(scipy.stats.entropy(target_trans, nb_freq))

                if 0 > (dist_before - dist_after):
                    #Reject the swap, so swap back.
                    p = math.exp((dist_before - dist_after) / T)
                    r = np.random.uniform(low=0, high=1)
                    if p < r:
                        celltype_assignment, nb_count, onehot_ct = before_ct_assign, before_nb_count, before_onehot_ct
                    else:
                        dist_before = dist_after
                else:
                    dist_before = dist_after
                iter_n += 1
        #     print(dist_before - dist_after)
            T *= decay
            if iter_n % 100 == 0:
                error = np.linalg.norm(nb_freq - target_trans)
                error_record.append(error)
            if iter_n % 1000 ==0:
                print("%5d iteration, error %.3f" % (iter_n, error))

            if len(error_record) > same_errlen:
                if len(set(error_record[-same_errlen:])) == 1:
                    criterion3 = False
    else:
        max_iter = bigsample_max_iter
        criterion3 = True
        while error > tol and iter_n < max_iter:

            for i in range(chain_len):

                #dist_before = sum(scipy.stats.entropy(target_trans,nb_freq))

                before_ct_assign, before_onehot_ct, before_nb_count = celltype_assignment.copy(), onehot_ct.copy(), nb_count.copy()

                swap_i, swap_j = swap_ct(
                    celltype_assignment=celltype_assignment,
                    Num_celltype=Num_celltype,
                    swap_num=swap_num)

                celltype_assignment[swap_i[0]], celltype_assignment[
                    swap_j[0]] = celltype_assignment[
                        swap_j[0]], celltype_assignment[swap_i[0]]
                onehot_ct[swap_i[0]], onehot_ct[swap_j[0]] = onehot_ct[
                    swap_j[0]], onehot_ct[swap_i[0]]

                nb_count = get_swap_nb_count(nb_count=nb_count,
                                             swap_i=swap_i,
                                             swap_j=swap_j,
                                             sn=sn)
                # nb_count = sn * onehot_ct
                nb_freq = get_nb_freq(nb_count=nb_count, onehot_ct=onehot_ct)

                dist_after = sum(scipy.stats.entropy(target_trans, nb_freq))

                if 0 > (dist_before - dist_after):
                    p = math.exp((dist_before - dist_after) / T)
                    r = np.random.uniform(low=0, high=1)
                    if p < r:
                        celltype_assignment, nb_count, onehot_ct = before_ct_assign, before_nb_count, before_onehot_ct
                    else:
                        dist_before = dist_after
                else:
                    dist_before = dist_after
                iter_n += 1
        #     print(dist_before - dist_after)
            T *= decay
            if iter_n % 500 == 0:
                error = np.linalg.norm(nb_freq - target_trans)
                print("%5d iteration, error %.3f" % (iter_n, error))

    return celltype_assignment



import io
import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


# Package meta-data.
NAME = 'st-spider'
DESCRIPTION = 'A tools to simulate spatial transcriptomics data.'
EMAIL = '599568651@qq.com'
URL="https://github.com/YANG-ERA/Spider/tree/main"
AUTHOR ='Jiyuan Yang'
VERSION = '0.2.2'

setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    author_email=EMAIL,
	license='MIT',
    description=DESCRIPTION,
	url=URL,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=[
        "anndata==0.10.5.post1",
        "matplotlib==3.8.3",
        "numba==0.59.0",
        "numpy==1.23.4",
        "pandas==2.2.1",
        "scanpy==1.9.8",
        "scikit-learn==1.4.1.post1",
        "scipy==1.12.0",
        "seaborn==0.13.2",
        "squidpy==1.4.1",
        "torch==2.2.1",
    ]
)



#from .scsim import *
from .scsim import *
from .sim_expr import *
from .simulate_10X import *
from .Annealing import *
from .sim_naive import *
from .random_based_utils import *
from .enhance import  *

#from .joint_simulator import *
#from .opt import *
#from .data_op import *
#from .random_generator import *


